# This file is part of jack_mixer
#
# Copyright (C) 2006 Nedko Arnaudov <nedko@arnaudov.name>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import math

from ._jack_mixer import Scale


log = logging.getLogger(__name__)


class Mark:
    """Encapsulates scale linear function edge and coefficients for scale = a * dB + b formula"""

    def __init__(self, db, scale, text=None):
        self.db = db
        self.scale = scale
        if text is None:
            # TODO: l10n
            self.text = "%.0f" % math.fabs(db)
        else:
            self.text = text


class Base:
    """Scale abstraction, various scale implementation derive from this class"""

    def __init__(self, scale_id, description):
        self.marks = []
        self.scale_id = scale_id
        self.description = description
        self.scale = Scale()

    def add_threshold(self, db, scale, is_mark, text=None):
        self.scale.add_threshold(db, scale)
        if is_mark:
            self.marks.append(Mark(db, scale, text))

    def calculate_coefficients(self):
        self.scale.calculate_coefficients()

    def db_to_scale(self, db):
        """Convert dBFS value to number in range 0.0-1.0 used in GUI"""
        # log.debug("db_to_scale(%f)", db)
        return self.scale.db_to_scale(db)

    def scale_to_db(self, scale):
        """Convert number in range 0.0-1.0 used in GUI to dBFS value"""
        return self.scale.scale_to_db(scale)

    def add_mark(self, db):
        self.marks.append(Mark(db, -1.0))

    def get_marks(self):
        return self.marks

    def scale_marks(self):
        for i in self.marks:
            if i.scale == -1.0:
                i.scale = self.db_to_scale(i.db)


# IEC 60268-18 Peak programme level meters - Digital audio peak level meter
# Adapted from meterbridge, may be wrong, I'm not buying standards, even if they cost $45
# If someone has the standard, please either share it with me or fix the code.
class IEC268(Base):
    """IEC 60268-18 Peak programme level meters - Digital audio peak level meter"""

    def __init__(self):
        Base.__init__(
            self,
            "iec_268",
            _("IEC 60268-18 Peak programme level meters - Digital audio peak level meter"),
        )
        self.add_threshold(-70.0, 0.0, False)
        self.add_threshold(-60.0, 0.05, True)
        self.add_threshold(-50.0, 0.075, True)
        self.add_threshold(-40.0, 0.15, True)
        self.add_mark(-35.0)
        self.add_threshold(-30.0, 0.3, True)
        self.add_mark(-25.0)
        self.add_threshold(-20.0, 0.5, True)
        self.add_mark(-15.0)
        self.add_mark(-10.0)
        self.add_mark(-5.0)
        self.add_threshold(0.0, 1.0, True)
        self.calculate_coefficients()
        self.scale_marks()


class IEC268Minimalistic(Base):
    """IEC 60268-18 Peak programme level meters - Digital audio peak level meter,
    fewer marks"""

    def __init__(self):
        Base.__init__(
            self,
            "iec_268_minimalistic",
            _(
                "IEC 60268-18 Peak programme level meters - "
                "Digital audio peak level meter, fewer marks"
            ),
        )
        self.add_threshold(-70.0, 0.0, False)
        self.add_threshold(-60.0, 0.05, True)
        self.add_threshold(-50.0, 0.075, False)
        self.add_threshold(-40.0, 0.15, True)
        self.add_threshold(-30.0, 0.3, True)
        self.add_threshold(-20.0, 0.5, True)
        self.add_mark(-10.0)
        self.add_threshold(0.0, 1.0, True)
        self.calculate_coefficients()
        self.scale_marks()


class Linear70dB(Base):
    """Linear scale with range from -70 to 0 dBFS"""

    def __init__(self):
        Base.__init__(self, "linear_70dB", _("Linear scale with range from -70 to 0 dBFS"))
        self.add_threshold(-70.0, 0.0, False)
        self.add_mark(-60.0)
        self.add_mark(-50.0)
        self.add_mark(-40.0)
        self.add_mark(-35.0)
        self.add_mark(-30.0)
        self.add_mark(-25.0)
        self.add_mark(-20.0)
        self.add_mark(-15.0)
        self.add_mark(-10.0)
        self.add_mark(-5.0)
        self.add_threshold(0.0, 1.0, True)
        self.calculate_coefficients()
        self.scale_marks()


class Linear30dB(Base):
    """Linear scale with range from -30 to +30 dBFS"""

    def __init__(self):
        Base.__init__(self, "linear_30dB", _("Linear scale with range from -30 to +30 dBFS"))
        self.add_threshold(-30.0, 0.0, False)
        self.add_threshold(+30.0, 1.0, True)
        self.calculate_coefficients()
        self.scale_marks()


class K20(Base):
    """K20 scale"""

    def __init__(self):
        Base.__init__(self, "K20", _("K20 scale"))
        self.add_mark(-200, "")
        self.add_mark(-60, "-40")
        self.add_mark(-55, "")
        self.add_mark(-50, "-30")
        self.add_mark(-45, "")
        self.add_mark(-40, "-20")
        self.add_mark(-35, "")
        self.add_mark(-30, "-10")
        self.add_mark(-26, "-6")
        self.add_mark(-23, "-3")
        self.add_mark(-20, "0")
        self.add_mark(-17, "3")
        self.add_mark(-14, "6")
        self.add_mark(-10, "10")
        self.add_mark(-5, "15")
        self.add_mark(0, "20")

    def db_to_scale(self, db):
        v = math.pow(10.0, db / 20.0)
        return self.mapk20(v) / 450.0

    def add_mark(self, db, text):
        self.marks.append(Mark(db, self.db_to_scale(db), text))

    def mapk20(self, v):
        if v < 0.001:
            return 24000 * v
        v = math.log10(v) + 3
        if v < 2.0:
            return 24.3 + v * (100 + v * 16)
        if v > 3.0:
            v = 3.0
        return v * 161.7 - 35.1


class K14(Base):
    """K14 scale"""

    def __init__(self):
        Base.__init__(self, "K14", _("K14 scale"))
        self.add_mark(-200, "")
        self.add_mark(-54, "-40")
        self.add_mark(-35 - 14, "")
        self.add_mark(-30 - 14, "-30")
        self.add_mark(-25 - 14, "")
        self.add_mark(-20 - 14, "-20")
        self.add_mark(-15 - 14, "")
        self.add_mark(-10 - 14, "-10")
        self.add_mark(-6 - 14, "-6")
        self.add_mark(-3 - 14, "-3")
        self.add_mark(-14, "0")
        self.add_mark(3 - 14, "3")
        self.add_mark(6 - 14, "6")
        self.add_mark(10 - 14, "10")
        self.add_mark(14 - 14, "14")

    def db_to_scale(self, db):
        v = math.pow(10.0, db / 20.0)
        return self.mapk14(v) / 448.3

    def add_mark(self, db, text):
        self.marks.append(Mark(db, self.db_to_scale(db), text))

    def mapk14(self, v):
        if v < 0.002:
            return int(12000 * v)
        v = math.log10(v) + 2.7
        if v < 2.0:
            return int(20.3 + v * (80 + v * 32))
        if v > 2.7:
            v = 2.7
        return int(v * 200 - 91.7)


def scale_test1(scale):
    for i in range(-97 * 2, 1, 1):
        db = float(i) / 2.0
        print("%5.1f dB maps to %f" % (db, scale.db_to_scale(db)))


def scale_test2(scale):
    for i in range(101):
        s = float(i) / 100.0
        print("%.2f maps to %.1f dB" % (s, scale.scale_to_db(s)))


def print_db_to_scale(scale, db):
    print("%-.1f dB maps to %f" % (db, scale.db_to_scale(db)))


def scale_test3(scale):
    print_db_to_scale(scale, +77.0)
    print_db_to_scale(scale, +7.0)
    print_db_to_scale(scale, 0.0)
    print_db_to_scale(scale, -107.0)


def _test(*args, **kwargs):
    scale = Linear30dB()
    scale_test1(scale)
    scale_test2(scale)
    scale_test3(scale)


if __name__ == "__main__":
    _test()



#!/usr/bin/env python3

import sysconfig
from compileall import compile_dir
from os import environ, path

prefix = environ.get('MESON_INSTALL_PREFIX', '/usr/local')
datadir = path.join(prefix, 'share')
destdir = environ.get('MESON_INSTALL_DESTDIR_PREFIX', '')

# Package managers set this so we don't need to run
if 'DESTDIR' not in environ:
    from subprocess import call
    print('Updating icon cache...')
    call(['gtk-update-icon-cache', '-qtf', path.join(datadir, 'icons', 'hicolor')])

print('Compiling Python module to bytecode...')
moduledir = sysconfig.get_path('purelib', vars={'base': destdir})
compile_dir(path.join(moduledir, 'jack_mixer'), optimize=1, stripdir=destdir)



#
# _jack_mixer.pxd
#
# cython: language_level=3
#

from libcpp cimport bool

cdef extern from "jack_mixer.h":
    # scale.h
    ctypedef void * jack_mixer_scale_t;

    cdef jack_mixer_scale_t scale_create()

    cdef bool scale_add_threshold(jack_mixer_scale_t scale, float db, float scale_value)
    cdef void scale_remove_thresholds(jack_mixer_scale_t scale)
    cdef void scale_calculate_coefficients(jack_mixer_scale_t scale)
    cdef double scale_db_to_scale(jack_mixer_scale_t scale, double db)
    cdef double scale_scale_to_db(jack_mixer_scale_t scale, double scale_value)
    cdef void scale_destroy(jack_mixer_scale_t scale)

    # jack_mixer.h
    ctypedef void * jack_mixer_t
    ctypedef void * jack_mixer_channel_t
    ctypedef void * jack_mixer_output_channel_t
    ctypedef void * jack_mixer_threshold_t

    ctypedef void (*midi_change_callback)(void *) noexcept with gil

    cdef enum midi_behavior_mode:
        pass

    cdef enum meter_mode:
        pass

    ctypedef enum jack_mixer_error_t:
        pass

    cdef jack_mixer_error_t jack_mixer_error();

    cdef const char* jack_mixer_error_str();

    # mixer
    cdef jack_mixer_t mixer_create "create" (const char * jack_client_name_ptr, bool stereo)
    cdef void mixer_destroy "destroy" (jack_mixer_t mixer)
    cdef unsigned int mixer_get_channels_count "get_channels_count" (jack_mixer_t mixer)
    cdef const char * mixer_get_client_name "get_client_name" (jack_mixer_t mixer)
    cdef int mixer_get_last_midi_cc "get_last_midi_cc" (jack_mixer_t mixer)
    cdef void mixer_set_last_midi_cc "set_last_midi_cc" (
        jack_mixer_t mixer,
        int new_channel)
    cdef int mixer_get_midi_behavior_mode "get_midi_behavior_mode" (jack_mixer_t mixer)
    cdef void mixer_set_midi_behavior_mode "set_midi_behavior_mode" (
        jack_mixer_t mixer,
        midi_behavior_mode mode)
    cdef jack_mixer_channel_t mixer_add_channel "add_channel" (
        jack_mixer_t mixer,
        const char * channel_name,
        bool stereo)
    cdef jack_mixer_output_channel_t mixer_add_output_channel "add_output_channel" (
        jack_mixer_t mixer,
        const char * channel_name,
        bool stereo,
        bool system)
    cdef bool mixer_get_kmetering "get_kmetering" (jack_mixer_t mixer)
    cdef void mixer_set_kmetering "set_kmetering" (jack_mixer_t mixer, bool flag)

    # not used by Python
    #cdef void channels_volumes_read(jack_mixer_t mixer)
    #cdef void remove_channels(jack_mixer_t mixer)

    # channel
    cdef const char * channel_get_name(jack_mixer_channel_t channel)
    cdef int channel_rename(jack_mixer_channel_t channel, const char * name)

    cdef double channel_abspeak_read(jack_mixer_channel_t channel, meter_mode mode)
    cdef void channel_abspeak_reset(jack_mixer_channel_t channel, meter_mode mode)

    cdef void channel_mono_meter_read(
        jack_mixer_channel_t channel,
        double * mono_ptr,
        meter_mode mode)
    cdef void channel_stereo_meter_read(
        jack_mixer_channel_t channel,
        double * left_ptr,
        double * right_ptr,
        meter_mode mode)

    cdef void channel_mono_kmeter_read(
        jack_mixer_channel_t channel,
        double * left_ptr,
        double * left_rms_ptr,
        meter_mode mode)
    cdef void channel_stereo_kmeter_read(
        jack_mixer_channel_t channel,
        double * left_ptr,
        double * right_ptr,
        double * left_rms_ptr,
        double * right_rms_ptr,
        meter_mode mode)

    cdef void channel_mono_kmeter_reset(jack_mixer_channel_t channel)
    cdef void channel_stereo_kmeter_reset(jack_mixer_channel_t channel)

    cdef void channel_volume_write(jack_mixer_channel_t channel, double volume)
    cdef double channel_volume_read(jack_mixer_channel_t channel)

    cdef double channel_balance_read(jack_mixer_channel_t channel)
    cdef void channel_balance_write(jack_mixer_channel_t channel, double balance)

    cdef bool channel_is_out_muted(jack_mixer_channel_t channel)
    cdef void channel_out_mute(jack_mixer_channel_t channel)
    cdef void channel_out_unmute(jack_mixer_channel_t channel)

    cdef bool channel_is_soloed(jack_mixer_channel_t channel)
    cdef void channel_solo(jack_mixer_channel_t channel)
    cdef void channel_unsolo(jack_mixer_channel_t channel)

    cdef bool channel_is_stereo(jack_mixer_channel_t channel)

    cdef void channel_set_midi_change_callback(
        jack_mixer_channel_t channel,
        midi_change_callback callback,
        void * user_data)
    cdef bool channel_get_midi_in_got_events(jack_mixer_channel_t channel)

    cdef int channel_autoset_balance_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_autoset_mute_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_autoset_solo_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_autoset_volume_midi_cc(jack_mixer_channel_t channel)

    cdef int channel_get_balance_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_get_mute_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_get_solo_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_get_volume_midi_cc(jack_mixer_channel_t channel)
    cdef int channel_set_balance_midi_cc(jack_mixer_channel_t channel, int new_cc)
    cdef int channel_set_mute_midi_cc(jack_mixer_channel_t channel, int new_cc)
    cdef int channel_set_solo_midi_cc(jack_mixer_channel_t channel, int new_cc)
    cdef int channel_set_volume_midi_cc(jack_mixer_channel_t channel, int new_cc)

    cdef void channel_set_midi_scale(jack_mixer_channel_t channel, jack_mixer_scale_t scale)

    cdef void channel_set_midi_cc_balance_picked_up(jack_mixer_channel_t channel, bool status)
    cdef void channel_set_midi_cc_volume_picked_up(jack_mixer_channel_t channel, bool status)

    cdef void remove_channel(jack_mixer_channel_t channel)

    # output channel
    cdef bool output_channel_is_muted(
        jack_mixer_output_channel_t output_channel,
        jack_mixer_channel_t channel)
    cdef void output_channel_set_muted(
        jack_mixer_output_channel_t output_channel,
        jack_mixer_channel_t channel,
        bool muted_value)
    cdef bool output_channel_is_prefader(
        jack_mixer_output_channel_t output_channel)
    cdef void output_channel_set_prefader(
        jack_mixer_output_channel_t output_channel,
        bool pfl_value)
    cdef bool output_channel_is_in_prefader(
        jack_mixer_output_channel_t output_channel,
        jack_mixer_channel_t input_channel)
    cdef void output_channel_set_in_prefader(
        jack_mixer_output_channel_t output_channel,
        jack_mixer_channel_t input_channel,
        bool prefader_value)
    cdef bool output_channel_is_solo(
        jack_mixer_output_channel_t output_channel,
        jack_mixer_channel_t channel)
    cdef void output_channel_set_solo(
        jack_mixer_output_channel_t output_channel,
        jack_mixer_channel_t channel,
        bool solo_value)
    cdef void remove_output_channel(jack_mixer_output_channel_t output_channel)



#!/usr/bin/env python
"""Create/Update man page(s) from ReST source file(s) to be included in source distribution."""

import argparse
import shutil
import sys
from os import chdir, environ, getcwd
from os.path import join

from subprocess import run

build_root = environ.get("MESON_BUILD_ROOT")
dist_root = environ.get("MESON_DIST_ROOT")
source_root = environ.get("MESON_SOURCE_ROOT")

ap = argparse.ArgumentParser()
ap.add_argument("-v", "--verbose", action="store_true", help="Be more verbose.")
ap.add_argument("man_page", nargs="*", help="Man page(s) to create.")
args = ap.parse_args()

if args.verbose:
    print("cwd:", getcwd())
    print("build root:", build_root)
    print("dist root:", dist_root)
    print("source root:", source_root)
    print("sys.argv:", sys.argv)

for man in args.man_page:
    target = join("docs", man)
    dst = join(dist_root, "docs", man)

    print("Creating man page '{}'".format(target))
    cmd = ["ninja"]

    if args.verbose:
        cmd += ["-v"]

    cmd += [target]

    chdir(build_root)
    proc = run(cmd)

    if proc.returncode != 0:
        sys.exit("'ninja' returned non-zero ({}) for target '{}'.".format(proc.returncode, target))

    shutil.copy(target, dst)



# This file is part of jack_mixer
#
# Copyright (C) 2006 Nedko Arnaudov <nedko@arnaudov.name>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging


log = logging.getLogger(__name__)


class SerializationBackend:
    """Base class for serialization backends"""

    def get_root_serialization_object(self, name):
        """Returns serialization object where properties of root object
        will be serialized to"""
        # this method should never be called for the base class
        raise NotImplementedError

    def get_child_serialization_object(self, name, backend_object):
        # this method should never be called for the base class
        raise NotImplementedError


class SerializationObjectBackend:
    """Base class for serialization backend objects where real object
    properties will be serialized to or unserialized from."""

    def add_property(self, name, value):
        """Serialize particular property"""
        pass

    def get_childs(self):
        pass

    def get_properties(self):
        pass

    def serialization_name(self):
        return None


class SerializedObject:
    """Base class for object supporting serialization"""

    def serialization_name(self):
        return None

    def serialize(self, object_backend):
        """Serialize properties of called object into supplied serialization_object_backend"""
        pass

    def serialization_get_childs(self):
        """Get child objects tha required and support serialization"""
        return []

    def unserialize_property(self, name, value):
        pass

    def unserialize_child(self, name):
        return None


class Serializator:
    def __init__(self):
        pass

    def serialize(self, root, backend):
        self.serialize_one(
            backend, root, backend.get_root_serialization_object(root.serialization_name())
        )

    def unserialize(self, root, backend):
        backend_object = backend.get_root_unserialization_object(root.serialization_name())
        if backend_object is None:
            return False

        return self.unserialize_one(backend, root, backend_object)

    def unserialize_one(self, backend, object, backend_object):
        log.debug("Unserializing %r.", object)
        properties = backend_object.get_properties()
        for name, value in properties.items():
            log.debug("Property %s = %s", name, value)
            if not object.unserialize_property(name, value):
                return False

        backend_childs = backend_object.get_childs()
        for backend_child in backend_childs:
            name = backend_child.serialization_name()
            child = object.unserialize_child(name)
            if not child:
                return False
            if not self.unserialize_one(backend, child, backend_child):
                return False

        return True

    def serialize_one(self, backend, object, backend_object):
        object.serialize(backend_object)
        childs = object.serialization_get_childs()
        for child in childs:
            log.debug("Serializing child %r.", child)
            self.serialize_one(
                backend,
                child,
                backend.get_child_serialization_object(child.serialization_name(), backend_object),
            )



# This file is part of jack_mixer
#
# Copyright (C) 2006 Nedko Arnaudov <nedko@arnaudov.name>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

import cairo
from gi.repository import Gtk
from gi.repository import Gdk

from . import scale as scalemod


log = logging.getLogger(__name__)

METER_MIN_WIDTH = 24
METER_MAX_WIDTH = 40

class MeterWidget(Gtk.DrawingArea):
    def __init__(self, scale):
        log.debug("Creating MeterWidget for scale %s", scale)
        super().__init__()
        self.color_bg = Gdk.Color(0, 0, 0)
        self.color_value = None
        self.color_mark = Gdk.Color(int(65535 * 0.2), int(65535 * 0.2), int(65535 * 0.2))
        self.width = 0
        self.height = 0
        self.cache_surface = None

        self.min_width = METER_MIN_WIDTH
        self.preferred_width = METER_MAX_WIDTH
        self.current_width_size_request = self.preferred_width
        self.preferred_height = 200
        self.set_scale(scale)

        self.widen()

        self.connect("draw", self.draw)
        self.connect("size-allocate", self.on_size_allocate)

    def narrow(self):
        return self.widen(False)

    def widen(self, flag=True):
        self.current_width_size_request = self.preferred_width if flag else self.min_width
        self.set_size_request(self.current_width_size_request, self.preferred_height)

    def set_color(self, color):
        self.color_value = color
        self.cache_surface = None
        self.invalidate_all()

    def on_expose(self, widget, event):
        cairo_ctx = widget.window.cairo_create()
        # set a clip region for the expose event
        cairo_ctx.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
        cairo_ctx.clip()

        self.draw(cairo_ctx)

        return False

    def on_size_allocate(self, widget, allocation):
        self.width = float(allocation.width)
        self.height = float(allocation.height)
        self.font_size = 10
        self.cache_surface = None

    def invalidate_all(self):
        self.queue_draw_area(0, 0, int(self.width), int(self.height))

    def draw_background(self, cairo_ctx):
        if not self.cache_surface:
            self.cache_surface = cairo.Surface.create_similar(
                cairo_ctx.get_target(), cairo.CONTENT_COLOR, int(self.width), int(self.height)
            )
            cache_cairo_ctx = cairo.Context(self.cache_surface)

            cache_cairo_ctx.set_source_rgba(0, 0, 0, 0)
            cache_cairo_ctx.rectangle(0, 0, self.width, self.height)
            cache_cairo_ctx.fill()

            cache_cairo_ctx.set_source_rgba(0.2, 0.2, 0.2, 1)
            cache_cairo_ctx.select_font_face("Fixed")
            cache_cairo_ctx.set_font_size(self.font_size)
            glyph_width = self.font_size * 3 / 5  # avarage glyph ratio

            for mark in self.scale.get_marks():
                mark_position = int(self.height * (1 - mark.scale))
                cache_cairo_ctx.move_to(0, mark_position)
                cache_cairo_ctx.line_to(self.width, mark_position)
                cache_cairo_ctx.stroke()
                x_correction = self.width / 2 - glyph_width * len(mark.text) / 2
                cache_cairo_ctx.move_to(x_correction, mark_position - 2)
                cache_cairo_ctx.show_text(mark.text)

        cairo_ctx.set_source_surface(self.cache_surface, 0, 0)
        cairo_ctx.paint()

    def draw_value(self, cairo_ctx, value, x, width):
        if self.color_value is not None:
            cairo_ctx.set_source_rgb(
                self.color_value.red / 65535.0,
                self.color_value.green / 65535.0,
                self.color_value.blue / 65535.0,
            )
        else:
            height = self.height
            gradient = cairo.LinearGradient(1, 1, width - 1, height - 1)

            if self.scale.scale_id == "K20":
                gradient.add_color_stop_rgb(0, 1, 0, 0)
                gradient.add_color_stop_rgb(0.38, 1, 1, 0)
                gradient.add_color_stop_rgb(0.5, 0, 1, 0)
                gradient.add_color_stop_rgb(1, 0, 0, 1)
            elif self.scale.scale_id == "K14":
                gradient.add_color_stop_rgb(0, 1, 0, 0)
                gradient.add_color_stop_rgb(1 - self.scale.db_to_scale(-14), 1, 1, 0)
                gradient.add_color_stop_rgb(1 - self.scale.db_to_scale(-24), 0, 1, 0)
                gradient.add_color_stop_rgb(1, 0, 0, 1)
            else:
                gradient.add_color_stop_rgb(0, 1, 0, 0)
                gradient.add_color_stop_rgb(0.2, 1, 1, 0)
                gradient.add_color_stop_rgb(1, 0, 1, 0)

            cairo_ctx.set_source(gradient)

        cairo_ctx.rectangle(x, self.height * (1 - value), width, self.height * value)
        cairo_ctx.fill()

    def draw_peak(self, cairo_ctx, value, x, width):
        cairo_ctx.set_source_rgb(1, 1, 1)
        cairo_ctx.rectangle(x, self.height * (1 - value), width, 2.5)
        cairo_ctx.fill()

    def set_scale(self, scale):
        self.kmetering = isinstance(scale, (scalemod.K14, scalemod.K20))
        self.scale = scale
        self.cache_surface = None
        self.invalidate_all()


class MonoMeterWidget(MeterWidget):
    def __init__(self, scale):
        super().__init__(scale)
        self.value = 0.0
        self.pk = 0.0
        self.raw_value = 0.0
        self.raw_pk = 0.0

    def draw(self, widget, cairo_ctx):
        self.draw_background(cairo_ctx)
        width = self.current_width_size_request / 1.5
        x = 0.5 * (self.width - width)
        self.draw_value(cairo_ctx, self.value, x, width)

        if self.kmetering:
            self.draw_peak(cairo_ctx, self.pk, x, width)

    def set_value(self, value):
        if value != self.raw_value:
            self.raw_value = value
            self.value = self.scale.db_to_scale(value)
            self.invalidate_all()

        if value == self.raw_value:
            return

        self.raw_value = value
        old_value = self.value
        self.value = self.scale.db_to_scale(value)

        if (abs(old_value-self.value) * self.height) > 1:
            self.invalidate_all()

    def set_value_kmeter(self, value, pk):
        if value == self.raw_value and pk == self.raw_pk:
            return

        self.raw_value = value
        self.raw_pk = pk
        old_value = self.value
        old_pk = self.pk
        self.value = self.scale.db_to_scale(value)
        self.pk = self.scale.db_to_scale(pk)

        if (abs(old_value - self.value) * self.height) > 0.01 or (
            abs(old_pk - self.pk) * self.height
        ) > 0.01:
            self.invalidate_all()


class StereoMeterWidget(MeterWidget):
    def __init__(self, scale):
        super().__init__(scale)
        self.pk_left = 0.0
        self.pk_right = 0.0

        self.left = 0.0
        self.right = 0.0

        self.raw_left_pk = 0.0
        self.raw_right_pk = 0.0

        self.raw_left = 0.0
        self.raw_right = 0.0

    def draw(self, widget, cairo_ctx):
        self.draw_background(cairo_ctx)

        width = self.current_width_size_request / 5.0
        left_x = 0.5 * self.width - self.current_width_size_request / 5.0 - 0.5 * width
        right_x = 0.5 * self.width + self.current_width_size_request / 5.0 - 0.5 * width

        self.draw_value(cairo_ctx, self.left, left_x, width)
        self.draw_value(cairo_ctx, self.right, right_x, width)

        if self.kmetering:
            self.draw_peak(cairo_ctx, self.pk_left, left_x, width)
            self.draw_peak(cairo_ctx, self.pk_right, right_x, width)

    def set_values(self, left, right):
        if left == self.raw_left and right == self.raw_right:
            return

        self.raw_left = left
        self.raw_right = right
        old_left = self.left
        old_right = self.right
        self.left = self.scale.db_to_scale(left)
        self.right = self.scale.db_to_scale(right)

        if (
            (abs(old_left - self.left) * self.height) > 0.01
            or (abs(old_right - self.right) * self.height) > 0.01
        ):
            self.invalidate_all()

    def set_values_kmeter(self, left, right, pk_l, pk_r):
        if (
            left == self.raw_left
            and right == self.raw_right
            and pk_l == self.raw_left_pk
            and pk_r == self.raw_right_pk
        ):
            return

        self.raw_left = left
        self.raw_right = right
        self.raw_left_pk = pk_l
        self.raw_right_pk = pk_r
        old_left = self.left
        old_right = self.right
        old_pk_left = self.pk_left
        old_pk_right = self.pk_right
        self.left = self.scale.db_to_scale(left)
        self.right = self.scale.db_to_scale(right)
        self.pk_left = self.scale.db_to_scale(pk_l)
        self.pk_right = self.scale.db_to_scale(pk_r)

        if (
            (abs(old_left - self.left) * self.height) > 0.01
            or (abs(old_right - self.right) * self.height) > 0.01
            or (abs(old_pk_left - self.pk_left) * self.height) > 0.01
            or (abs(old_pk_right - self.pk_right) * self.height) > 0.01
        ):
            self.invalidate_all()



#!/usr/bin/env python
#
# This file is part of jack_mixer
#
# Copyright (C) 2006 Nedko Arnaudov <nedko@arnaudov.name>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import jack_mixer_c

mixer = jack_mixer_c.Mixer("test")

print("Channels count: %u" % mixer.channels_count)
channel = mixer.add_channel("Channel 1", True)

if channel.is_stereo:
    channel_type = "Stereo"
else:
    channel_type = "Mono"

channel_name = channel.name

print('%s channel "%s"' % (channel_type, channel_name))

print("Channel meter read %s" % repr(channel.meter))
print("Channels count: %u" % mixer.channels_count)

channel.remove()

print("Channels count: %u" % mixer.channels_count)



#!/usr/bin/env python3
"""Compile translated message catalog *.po files to *.mo with msgfmt."""

import glob
import os
from os.path import basename, join, splitext
from subprocess import run

LOCALEDIR = join(os.getcwd(), "data", "locale")

for po in glob.glob(join(LOCALEDIR, "*.po")):
    fn = basename(po)
    domain, lang = splitext(fn)[0].rsplit("-", 1)
    langdir = join(LOCALEDIR, lang, "LC_MESSAGES")
    mo = join(langdir, domain + ".mo")
    print(f"Compiling {fn} to {mo} ...")
    os.makedirs(langdir, exist_ok=True)
    run(["msgfmt", "-o", mo, po])



#!/usr/bin/env python

import argparse
import shutil
import sys
from os import chdir, environ, getcwd
from os.path import join

from subprocess import run

build_root = environ.get("MESON_BUILD_ROOT")
dist_root = environ.get("MESON_DIST_ROOT")
source_root = environ.get("MESON_SOURCE_ROOT")

ap = argparse.ArgumentParser()
ap.add_argument("-v", "--verbose", action="store_true", help="Be more verbose.")
ap.add_argument("mod_source", nargs="*", help="Cython module C source target(s) (*.c).")
args = ap.parse_args()

if args.verbose:
    print("cwd:", getcwd())
    print("build root:", build_root)
    print("dist root:", dist_root)
    print("source root:", source_root)
    print("sys.argv:", sys.argv)

for mod in args.mod_source:
    target = join("src", mod)
    dst = join(dist_root, "src", mod)

    print("Updating Cython module C source '{}'...".format(target))
    cmd = ["ninja"]

    if args.verbose:
        cmd += ["-v"]

    cmd += [target]

    chdir(build_root)
    proc = run(cmd)

    if proc.returncode != 0:
        sys.exit("'ninja' returned non-zero ({}) for target '{}'.".format(proc.returncode, target))

    shutil.copy(target, dst)



#
# cython: language_level=3
#
"""Python bindings for jack_mixer.c and scale.c using Cython."""

__all__ = ("Scale", "MidiBehaviour", "Mixer")

import enum

from _jack_mixer cimport *


cdef void midi_change_callback_func(void *userdata) noexcept with gil:
    """Wrapper for a Python callback function for MIDI input."""
    channel = <object> userdata
    channel._midi_change_callback()


class MidiBehaviour(enum.IntEnum):
    """MIDI control behaviour.

    `JUMP_TO_VALUE`

    Received MIDI control messages affect mixer directly.

    `PICK_UP`

    Received MIDI control messages have to match up with current
    mixer value first (within a small margin), before further
    changes take effect.
    """
    JUMP_TO_VALUE = 0
    PICK_UP = 1


class MeterMode(enum.IntEnum):
    """Choose between pre-fader or post-fader metering.

    `PRE_FADER`

    Meter signal before applying fader.

    `POST_FADER`

    Meter signal after applying fader.
    """
    PRE_FADER = 0
    POST_FADER = 1


cdef class Scale:
    """Mixer level scale representation.

    Wraps `jack_mixer_scale_t` struct.
    """

    cdef jack_mixer_scale_t _scale

    def __cinit__(self):
        self._scale = scale_create()

    def __dealloc__(self):
        if self._scale:
            scale_destroy(self._scale)

    cpdef bool add_threshold(self, float db, float scale_value):
        """Add scale threshold."""
        return scale_add_threshold(self._scale, db, scale_value)

    cpdef void remove_thresholds(self):
        """Remove scale threshold."""
        scale_remove_thresholds(self._scale)

    cpdef void calculate_coefficients(self):
        """Calculate scale coefficients."""
        scale_calculate_coefficients(self._scale)

    cpdef double db_to_scale(self, double db):
        """Return scale value responding to given dB value."""
        return scale_db_to_scale(self._scale, db)

    cpdef double scale_to_db(self, double scale_value):
        """Return dB value responding to given scale value."""
        return scale_scale_to_db(self._scale, scale_value)


cdef class Mixer:
    """Jack Mixer representation.

    Wraps `jack_mixer_t` struct.
    """

    cdef jack_mixer_t _mixer
    cdef bool _stereo

    def __cinit__(self, name, stereo=True):
        self._stereo = stereo
        self._mixer = mixer_create(name.encode('utf-8'), stereo)
        if self._mixer == NULL:
            raise RuntimeError(jack_mixer_error_str().decode('utf-8'))

    def __dealloc__(self):
        if self._mixer:
            mixer_destroy(self._mixer)

    def destroy(self):
        """Close mixer Jack client and destroy mixer instance.

        The instance must not be used anymore after calling this
        method.
        """
        if self._mixer:
            mixer_destroy(self._mixer)

    @property
    def channels_count(self):
        """Number of mixer channels."""
        return mixer_get_channels_count(self._mixer)

    @property
    def client_name(self):
        """Jack client name of mixer."""
        return mixer_get_client_name(self._mixer).decode('utf-8')

    @property
    def last_midi_cc(self):
        """Last received MIDI control change message."""
        return mixer_get_last_midi_cc(self._mixer)

    @last_midi_cc.setter
    def last_midi_cc(self, int channel):
        mixer_set_last_midi_cc(self._mixer, channel)

    @property
    def midi_behavior_mode(self):
        """MIDI control change behaviour mode.

        See `MidiBehaviour` enum for more information.
        """
        return MidiBehaviour(mixer_get_midi_behavior_mode(self._mixer))

    @midi_behavior_mode.setter
    def midi_behavior_mode(self, mode):
        mixer_set_midi_behavior_mode(self._mixer,
                                     mode.value if isinstance(mode, MidiBehaviour) else mode)

    cpdef add_channel(self, channel_name, stereo=None):
        """Add a stereo or mono input channel with given name to the mixer.

        Returns a `Channel` instance when successful or `None` if channel
        creation failed.
        """
        cdef jack_mixer_channel_t chan_ptr
        if stereo is None:
            stereo = self._stereo

        chan_ptr = mixer_add_channel(self._mixer, channel_name.encode('utf-8'), stereo)
        if chan_ptr == NULL:
            return None

        return Channel.new(chan_ptr)

    cpdef add_output_channel(self, channel_name, stereo=None, system=False):
        """Add a stereo or mono output channel with given name to the mixer.

        Returns a `OutputChannel` instance when successful or `None` if
        channel creation failed.
        """
        cdef jack_mixer_output_channel_t chan_ptr

        if stereo is None:
            stereo = self._stereo

        chan_ptr = mixer_add_output_channel(self._mixer, channel_name.encode('utf-8'), stereo,
                                            system)

        if chan_ptr == NULL:
            return None

        return OutputChannel.new(chan_ptr)

    @property
    def kmetering(self):
        """Using kmeters."""
        return mixer_get_kmetering(self._mixer)

    @kmetering.setter
    def kmetering(self, bool flag):
        mixer_set_kmetering(self._mixer, flag)

cdef class Channel:
    """Jack Mixer (input) channel representation.

    Wraps `jack_mixer_channel_t` struct.
    """

    cdef jack_mixer_channel_t _channel
    cdef object _midi_change_callback

    def __init__(self):
        raise TypeError("Channel instances can only be created via Mixer.add_channel().")

    @staticmethod
    cdef Channel new(jack_mixer_channel_t chan_ptr):
        """Create a new Channel instance.

        A pointer to an initialized `jack_mixer_channel_t` struct must be
        passed in.

        This should not be called directly but only via `Mixer.add_channel()`.
        """
        cdef Channel channel = Channel.__new__(Channel)
        channel._channel = chan_ptr
        return channel

    @property
    def abspeak_postfader(self):
        """Absolute peak of channel meter.

        Set to `None` to reset the absolute peak to -inf.
        Trying to set it to any other value will raise a `ValueError`.
        """
        return channel_abspeak_read(self._channel, MeterMode.POST_FADER)

    @abspeak_postfader.setter
    def abspeak_postfader(self, reset):
        if reset is not None:
            raise ValueError("abspeak can only be reset (set to None)")
        channel_abspeak_reset(self._channel, MeterMode.POST_FADER)

    @property
    def abspeak_prefader(self):
        """Absolute peak of channel meter.

        Set to `None` to reset the absolute peak to -inf.
        Trying to set it to any other value will raise a `ValueError`.
        """
        return channel_abspeak_read(self._channel, MeterMode.PRE_FADER)

    @abspeak_prefader.setter
    def abspeak_prefader(self, reset):
        if reset is not None:
            raise ValueError("abspeak can only be reset (set to None)")
        channel_abspeak_reset(self._channel, MeterMode.PRE_FADER)

    @property
    def balance(self):
        """Channel balance property."""
        return channel_balance_read(self._channel)

    @balance.setter
    def balance(self, double bal):
        channel_balance_write(self._channel, bal)

    @property
    def is_stereo(self):
        """Is channel stereo or mono?"""
        return channel_is_stereo(self._channel)

    @property
    def kmeter_prefader(self):
        """Read channel prefader kmeter.

        If channel is stereo, return a four-item tupel with
        ``(rms_left, rms_right, peak_left, peak_right)`` value.
        If channel is mono, return a tow-item tupel with ``(rms, peak)`` value.
        """
        cdef double peak_left, peak_right, left_rms, right_rms

        if channel_is_stereo(self._channel):
            channel_stereo_kmeter_read(
                self._channel, &peak_left, &peak_right, &left_rms, &right_rms, MeterMode.PRE_FADER)
            return (left_rms, right_rms, peak_left, peak_right)
        else:
            channel_mono_kmeter_read(self._channel, &peak_left, &left_rms, MeterMode.PRE_FADER)
            return (left_rms, peak_left)

    @property
    def kmeter_postfader(self):
        """Read channel postfader kmeter.

        If channel is stereo, return a four-item tupel with
        ``(rms_left, rms_right, peak_left, peak_right)`` value.
        If channel is mono, return a tow-item tupel with ``(rms, peak)`` value.
        """
        cdef double peak_left, peak_right, left_rms, right_rms

        if channel_is_stereo(self._channel):
            channel_stereo_kmeter_read(
                self._channel, &peak_left, &peak_right, &left_rms, &right_rms, MeterMode.POST_FADER)
            return (left_rms, right_rms, peak_left, peak_right)
        else:
            channel_mono_kmeter_read(self._channel, &peak_left, &left_rms, MeterMode.POST_FADER)
            return (left_rms, peak_left)

    def kmeter_reset(self):
        """Reset channel kmeters"""
        if channel_is_stereo(self._channel):
            channel_stereo_kmeter_reset(self._channel)
        else:
            channel_mono_kmeter_reset(self._channel)

    @property
    def meter_prefader(self):
        """Read channel meter.

        If channel is stereo, return a two-item tupel with (left, right) value.
        If channel is mono, return a tupel with the value as the only item.
        """
        cdef double left, right

        if channel_is_stereo(self._channel):
            channel_stereo_meter_read(self._channel, &left, &right, MeterMode.PRE_FADER)
            return (left, right)
        else:
            channel_mono_meter_read(self._channel, &left, MeterMode.PRE_FADER)
            return (left,)

    @property
    def meter_postfader(self):
        """Read channel meter.

        If channel is stereo, return a two-item tupel with (left, right) value.
        If channel is mono, return a tupel with the value as the only item.
        """
        cdef double left, right

        if channel_is_stereo(self._channel):
            channel_stereo_meter_read(self._channel, &left, &right, MeterMode.POST_FADER)
            return (left, right)
        else:
            channel_mono_meter_read(self._channel, &left, MeterMode.POST_FADER)
            return (left,)

    @property
    def midi_change_callback(self):
        """Function to be called when a channel property is changed via MIDI.

        The callback function takes no arguments.

        Assign `None` to remove any existing callback.
        """
        return self._midi_change_callback

    @midi_change_callback.setter
    def midi_change_callback(self, callback):
        self._midi_change_callback = callback
        if callback is None:
            channel_set_midi_change_callback(self._channel, NULL, NULL)
        else:
            channel_set_midi_change_callback(self._channel,
                                             &midi_change_callback_func,
                                             <void *>self)

    @property
    def name(self):
        """Channel name property."""
        return channel_get_name(self._channel).decode('utf-8')

    @name.setter
    def name(self, newname):
        if channel_rename(self._channel, newname.encode('utf-8')) != 0:
            raise RuntimeError(jack_mixer_error_str().decode('utf-8'))

    @property
    def out_mute(self):
        """Channel solo status property."""
        return channel_is_out_muted(self._channel)

    @out_mute.setter
    def out_mute(self, bool value):
        if value:
            channel_out_mute(self._channel)
        else:
            channel_out_unmute(self._channel)

    @property
    def solo(self):
        """Channel solo status property."""
        return channel_is_soloed(self._channel)

    @solo.setter
    def solo(self, bool value):
        if value:
            channel_solo(self._channel)
        else:
            channel_unsolo(self._channel)

    @property
    def midi_in_got_events(self):
        """Did channel receive any MIDI events assigned to one of its controls?

        Reading this property also resets it to False.
        """
        return channel_get_midi_in_got_events(self._channel)

    @property
    def midi_scale(self):
        """MIDI scale used by channel."""
        raise AttributeError("midi_scale can only be set.")

    @midi_scale.setter
    def midi_scale(self, Scale scale):
        channel_set_midi_scale(self._channel, scale._scale)

    @property
    def volume(self):
        """Channel volume property."""
        return channel_volume_read(self._channel)

    @volume.setter
    def volume(self, double vol):
        channel_volume_write(self._channel, vol)

    @property
    def balance_midi_cc(self):
        """MIDI CC assigned to control channel balance."""
        return channel_get_balance_midi_cc(self._channel)

    @balance_midi_cc.setter
    def balance_midi_cc(self, int cc):
        if channel_set_balance_midi_cc(self._channel, cc) != 0:
            raise ValueError(jack_mixer_error_str().decode('utf-8'))

    @property
    def mute_midi_cc(self):
        """MIDI CC assigned to control channel mute status."""
        return channel_get_mute_midi_cc(self._channel)

    @mute_midi_cc.setter
    def mute_midi_cc(self, int cc):
        if channel_set_mute_midi_cc(self._channel, cc) != 0:
            raise ValueError(jack_mixer_error_str().decode('utf-8'))

    @property
    def solo_midi_cc(self):
        """MIDI CC assigned to control channel solo status."""
        return channel_get_solo_midi_cc(self._channel)

    @solo_midi_cc.setter
    def solo_midi_cc(self, int cc):
        if channel_set_solo_midi_cc(self._channel, cc) != 0:
            raise ValueError(jack_mixer_error_str().decode('utf-8'))

    @property
    def volume_midi_cc(self):
        """MIDI CC assigned to control channel volume."""
        return channel_get_volume_midi_cc(self._channel)

    @volume_midi_cc.setter
    def volume_midi_cc(self, int cc):
        if channel_set_volume_midi_cc(self._channel, cc) != 0:
            raise ValueError(jack_mixer_error_str().decode('utf-8'))

    def autoset_balance_midi_cc(self):
        """Auto assign MIDI CC for channel balance."""
        return channel_autoset_balance_midi_cc(self._channel)

    def autoset_mute_midi_cc(self):
        """Auto assign MIDI CC for channel mute status."""
        return channel_autoset_mute_midi_cc(self._channel)

    def autoset_solo_midi_cc(self):
        """Auto assign MIDI CC for channel solo status."""
        return channel_autoset_solo_midi_cc(self._channel)

    def autoset_volume_midi_cc(self):
        """Auto assign MIDI CC for channel volume."""
        return channel_autoset_volume_midi_cc(self._channel)

    def remove(self):
        """Remove channel."""
        remove_channel(self._channel)

    def set_midi_cc_balance_picked_up(self, bool status):
        """Set whether balance value is out-of-sync with MIDI control."""
        channel_set_midi_cc_balance_picked_up(self._channel, status)

    def set_midi_cc_volume_picked_up(self, bool status):
        """Set whether volume value is out-of-sync with MIDI control."""
        channel_set_midi_cc_volume_picked_up(self._channel, status)



cdef class OutputChannel(Channel):
    """Jack Mixer output channel representation.

    Wraps `jack_mixer_output_channel_t` struct.

    Inherits from `Channel` class.
    """

    cdef jack_mixer_output_channel_t _output_channel

    def __init__(self):
        raise TypeError("OutputChannel instances can only be created via "
                        "Mixer.add_output_channel().")

    @staticmethod
    cdef OutputChannel new(jack_mixer_output_channel_t chan_ptr):
        """Create a new OutputChannel instance.

        A pointer to an initialzed `jack_mixer_output_channel_t` struct must
        be passed in.

        This should not be called directly but only via
        `Mixer.add_output_channel()`.
        """
        cdef OutputChannel channel = OutputChannel.__new__(OutputChannel)
        channel._output_channel = chan_ptr
        channel._channel = <jack_mixer_channel_t> chan_ptr
        return channel

    @property
    def prefader(self):
        return output_channel_is_prefader(self._output_channel)

    @prefader.setter
    def prefader(self, bool pfl):
        output_channel_set_prefader(self._output_channel, pfl)

    def is_in_prefader(self, Channel channel):
        """Is a channel set as prefader?"""
        return output_channel_is_in_prefader(self._output_channel, channel._channel)

    def set_in_prefader(self, Channel channel, bool value):
        """Set a channel as prefader."""
        output_channel_set_in_prefader(self._output_channel, channel._channel, value)

    def is_muted(self, Channel channel):
        """Is a channel set as muted?"""
        return output_channel_is_muted(self._output_channel, channel._channel)

    def set_muted(self, Channel channel, bool value):
        """Set a channel as muted."""
        output_channel_set_muted(self._output_channel, channel._channel, value)

    def is_solo(self, Channel channel):
        """Is a channel set as solo?"""
        return output_channel_is_solo(self._output_channel, channel._channel)

    def set_solo(self, Channel channel, bool value):
        """Set a channel as solo."""
        output_channel_set_solo(self._output_channel, channel._channel, value)

    def remove(self):
        """Remove output channel."""
        remove_output_channel(self._output_channel)



#!/usr/bin/env python3
"""Merge new/updated messages from jack_mixer.pot into existing translations with msgmerge."""

import glob
import os
from os.path import basename, join, splitext
from subprocess import run

LOCALEDIR = join(os.getcwd(), "data", "locale")

for po in glob.glob(join(LOCALEDIR, "*.po")):
    fn = basename(po)
    domain, lang = splitext(fn)[0].rsplit("-", 1)
    pot = join(LOCALEDIR, domain + ".pot")
    print(f"Merging new/updated messages from {pot} into {po} ...")
    run(["msgmerge", "-U", po, pot])



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
PyNSMClient -  A New Session Manager Client-Library in one file.

The Non-Session-Manager by Jonathan Moore Liles <male@tuxfamily.org>: http://non.tuxfamily.org/nsm/
New Session Manager by Nils Hilbricht et al  https://new-session-manager.jackaudio.org
With help from code fragments from https://github.com/attwad/python-osc ( DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE v2 )

MIT License

Copyright (c) since 2014: Laborejo Software Suite <info@laborejo.org>, All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging;
logger: logging.Logger  #filled by init with client logger.

import struct
import socket
from os import getenv, getpid, kill
import os
import os.path
import shutil
from uuid import uuid4
from sys import argv
from signal import signal, SIGTERM, SIGINT, SIGKILL #react to exit signals to close the client gracefully. Or kill if the client fails to do so.
from urllib.parse import urlparse

class _IncomingMessage(object):
    """Representation of a parsed datagram representing an OSC message.

    An OSC message consists of an OSC Address Pattern followed by an OSC
    Type Tag String followed by zero or more OSC Arguments.
    """

    def __init__(self, dgram):
        #NSM Broadcasts are bundles, but very simple ones. We only need to care about the single message it contains.
        #Therefore we can strip the bundle prefix and handle it as normal message.
        if b"#bundle" in dgram:
            bundlePrefix, singleMessage = dgram.split(b"/", maxsplit=1)
            dgram = b"/" + singleMessage  # / eaten by split
            self.isBroadcast = True
        else:
            self.isBroadcast = False
        self.LENGTH = 4 #32 bit
        self._dgram = dgram
        self._parameters = []
        self.parse_datagram()


    def get_int(self, dgram, start_index):
        """Get a 32-bit big-endian two's complement integer from the datagram.

        Args:
        dgram: A datagram packet.
        start_index: An index where the integer starts in the datagram.

        Returns:
        A tuple containing the integer and the new end index.

        Raises:
        ValueError if the datagram could not be parsed.
        """
        try:
            if len(dgram[start_index:]) < self.LENGTH:
                raise ValueError('Datagram is too short')
            return (
                struct.unpack('>i', dgram[start_index:start_index + self.LENGTH])[0], start_index + self.LENGTH)
        except (struct.error, TypeError) as e:
            raise ValueError('Could not parse datagram %s' % e)

    def get_string(self, dgram, start_index):
        """Get a python string from the datagram, starting at pos start_index.

        We receive always the full string, but handle only the part from the start_index internally.
        In the end return the offset so it can be added to the index for the next parameter.
        Each subsequent call handles less of the same string, starting further to the right.

        According to the specifications, a string is:
        "A sequence of non-null ASCII characters followed by a null,
        followed by 0-3 additional null characters to make the total number
        of bits a multiple of 32".

        Args:
        dgram: A datagram packet.
        start_index: An index where the string starts in the datagram.

        Returns:
        A tuple containing the string and the new end index.

        Raises:
        ValueError if the datagram could not be parsed.
        """
        #First test for empty string, which is nothing, followed by a terminating \x00 padded by three additional \x00.
        if dgram[start_index:].startswith(b"\x00\x00\x00\x00"):
            return "", start_index + 4

        #Otherwise we have a non-empty string that must follow the rules of the docstring.

        offset = 0
        try:
            while dgram[start_index + offset] != 0:
                offset += 1
            if offset == 0:
                raise ValueError('OSC string cannot begin with a null byte: %s' % dgram[start_index:])
            # Align to a byte word.
            if (offset) % self.LENGTH == 0:
                offset += self.LENGTH
            else:
                offset += (-offset % self.LENGTH)
            # Python slices do not raise an IndexError past the last index,
                # do it ourselves.
            if offset > len(dgram[start_index:]):
                raise ValueError('Datagram is too short')
            data_str = dgram[start_index:start_index + offset]
            return data_str.replace(b'\x00', b'').decode('utf-8'), start_index + offset
        except IndexError as ie:
            raise ValueError('Could not parse datagram %s' % ie)
        except TypeError as te:
            raise ValueError('Could not parse datagram %s' % te)

    def get_float(self, dgram, start_index):
        """Get a 32-bit big-endian IEEE 754 floating point number from the datagram.

          Args:
            dgram: A datagram packet.
            start_index: An index where the float starts in the datagram.

          Returns:
            A tuple containing the float and the new end index.

          Raises:
            ValueError if the datagram could not be parsed.
        """
        try:
            return (struct.unpack('>f', dgram[start_index:start_index + self.LENGTH])[0], start_index + self.LENGTH)
        except (struct.error, TypeError) as e:
            raise ValueError('Could not parse datagram %s' % e)

    def parse_datagram(self):
        try:
            self._address_regexp, index = self.get_string(self._dgram, 0)
            if not self._dgram[index:]:
                # No params is legit, just return now.
                return

            # Get the parameters types.
            type_tag, index = self.get_string(self._dgram, index)
            if type_tag.startswith(','):
                type_tag = type_tag[1:]

            # Parse each parameter given its type.
            for param in type_tag:
                if param == "i":  # Integer.
                    val, index = self.get_int(self._dgram, index)
                elif param == "f":  # Float.
                    val, index = self.get_float(self._dgram, index)
                elif param == "s":  # String.
                    val, index = self.get_string(self._dgram, index)
                else:
                    logger.warning("Unhandled parameter type: {0}".format(param))
                    continue
                self._parameters.append(val)
        except ValueError as pe:
            #raise ValueError('Found incorrect datagram, ignoring it', pe)
            # Raising an error is not ignoring it!
            logger.warning("Found incorrect datagram, ignoring it. {}".format(pe))

    @property
    def oscpath(self):
        """Returns the OSC address regular expression."""
        return self._address_regexp

    @staticmethod
    def dgram_is_message(dgram):
        """Returns whether this datagram starts as an OSC message."""
        return dgram.startswith(b'/')

    @property
    def size(self):
        """Returns the length of the datagram for this message."""
        return len(self._dgram)

    @property
    def dgram(self):
        """Returns the datagram from which this message was built."""
        return self._dgram

    @property
    def params(self):
        """Convenience method for list(self) to get the list of parameters."""
        return list(self)

    def __iter__(self):
        """Returns an iterator over the parameters of this message."""
        return iter(self._parameters)

class _OutgoingMessage(object):
    def __init__(self, oscpath):
        self.LENGTH = 4 #32 bit
        self.oscpath = oscpath
        self._args = []

    def write_string(self, val):
        dgram = val.encode('utf-8')
        diff = self.LENGTH - (len(dgram) % self.LENGTH)
        dgram += (b'\x00' * diff)
        return dgram

    def write_int(self, val):
        return struct.pack('>i', val)

    def write_float(self, val):
        return struct.pack('>f', val)

    def add_arg(self, argument):
        t = {str:"s", int:"i", float:"f"}[type(argument)]
        self._args.append((t, argument))

    def build(self):
        dgram = b''

        #OSC Path
        dgram += self.write_string(self.oscpath)

        if not self._args:
            dgram += self.write_string(',')
            return dgram

        # Write the parameters.
        arg_types = "".join([arg[0] for arg in self._args])
        dgram += self.write_string(',' + arg_types)
        for arg_type, value in self._args:
            f = {"s":self.write_string, "i":self.write_int, "f":self.write_float}[arg_type]
            dgram += f(value)
        return dgram

class NSMNotRunningError(Exception):
    """Error raised when environment variable $NSM_URL was not found."""

class NSMClient(object):
    """The representation of the host programs as NSM sees it.
    Technically consists of an udp server and a udp client.

    Does not run an event loop itself and depends on the host loop.
    E.g. a Qt timer or just a simple while True: sleep(0.1) in Python."""
    def __init__(self, prettyName, supportsSaveStatus, saveCallback, openOrNewCallback, exitProgramCallback, hideGUICallback=None, showGUICallback=None, broadcastCallback=None, sessionIsLoadedCallback=None, loggingLevel = "info"):

        self.nsmOSCUrl = self.getNsmOSCUrl() #this fails and raises NSMNotRunningError if NSM is not available. Host programs can ignore it or exit their program.

        self.realClient = True
        self.cachedSaveStatus = None #save status checks for this.

        global logger
        logger = logging.getLogger(prettyName)
        logger.info("import")
        if loggingLevel == "info" or loggingLevel == 20:
            logging.basicConfig(level=logging.INFO) #development
            logger.info("Starting PyNSM2 Client with logging level INFO. Switch to 'error' for a release!") #the NSM name is not ready yet so we just use the pretty name
        elif loggingLevel == "error" or loggingLevel == 40:
            logging.basicConfig(level=logging.ERROR) #production
        else:
            logging.warning("Unknown logging level: {}. Choose 'info' or 'error'".format(loggingLevel))
            logging.basicConfig(level=logging.INFO) #development

        #given parameters,
        self.prettyName = prettyName #keep this consistent! Settle for one name.
        self.supportsSaveStatus = supportsSaveStatus
        self.saveCallback = saveCallback
        self.exitProgramCallback = exitProgramCallback
        self.openOrNewCallback = openOrNewCallback #The host needs to: Create a jack client with ourClientNameUnderNSM - Open the saved file and all its resources
        self.broadcastCallback = broadcastCallback
        self.hideGUICallback = hideGUICallback
        self.showGUICallback = showGUICallback
        self.sessionIsLoadedCallback = sessionIsLoadedCallback

        #Reactions get the raw _IncomingMessage OSC object
        #A client can add to reactions.
        self.reactions = {
                          "/nsm/client/save" : self._saveCallback,
                          "/nsm/client/show_optional_gui" : lambda msg: self.showGUICallback(),
                          "/nsm/client/hide_optional_gui" : lambda msg: self.hideGUICallback(),
                          "/nsm/client/session_is_loaded" : self._sessionIsLoadedCallback,
                          #Hello source-code reader. You can add your own reactions here by nsmClient.reactions[oscpath]=func, where func gets the raw _IncomingMessage OSC object as argument.
                          #broadcast is handled directly by the function because it has more parameters
                          }
        #self.discardReactions = set(["/nsm/client/session_is_loaded"])
        self.discardReactions = set()

        #Networking and Init
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) #internet, udp
        self.sock.bind(('', 0)) #pick a free port on localhost.
        ip, port = self.sock.getsockname()
        self.ourOscUrl = f"osc.udp://{ip}:{port}/"

        self.executableName = self.getExecutableName()

        #UNIX Signals. Used for quit.
        signal(SIGTERM, self.sigtermHandler) #NSM sends only SIGTERM. #TODO: really? pynsm version 1 handled sigkill as well.
        signal(SIGINT, self.sigtermHandler)

        #The following instance parameters are all set in announceOurselves
        self.serverFeatures = None
        self.sessionName = None
        self.ourPath = None
        self.ourClientNameUnderNSM = None
        self.ourClientId = None # the "file extension" of ourClientNameUnderNSM
        self.isVisible = None #set in announceGuiVisibility
        self.saveStatus = True # true is clean. false means we need saving.

        self.announceOurselves()

        assert self.serverFeatures, self.serverFeatures
        assert self.sessionName, self.sessionName
        assert self.ourPath, self.ourPath
        assert self.ourClientNameUnderNSM, self.ourClientNameUnderNSM

        self.sock.setblocking(False) #We have waited for tha handshake. Now switch blocking off because we expect sock.recvfrom to be empty in 99.99...% of the time so we shouldn't wait for the answer.
        #After this point the host must include self.reactToMessage in its event loop

        #We assume we are save at startup.
        self.announceSaveStatus(isClean = True)

        logger.info("NSMClient client init complete. Going into listening mode.")


    def reactToMessage(self):
        """This is the main loop message. It is added to the clients event loop."""
        try:
            data, addr = self.sock.recvfrom(4096) #4096 is quite big. We don't expect nsm messages this big. Better safe than sorry. However, messages will crash the program if they are bigger than 4096.
        except BlockingIOError: #happens while no data is received. Has nothing to do with blocking or not.
            return None

        msg = _IncomingMessage(data)
        if msg.oscpath in self.reactions:
            self.reactions[msg.oscpath](msg)
        elif msg.oscpath in self.discardReactions:
            pass
        elif msg.oscpath == "/reply" and msg.params == ["/nsm/server/open", "Loaded."]: #NSM sends that all programs of the session were loaded.
            logger.info ("Got /reply Loaded from NSM Server")
        elif msg.oscpath == "/reply" and msg.params == ["/nsm/server/save", "Saved."]: #NSM sends that all program-states are saved. Does only happen from the general save instruction, not when saving our client individually
            logger.info ("Got /reply Saved from NSM Server")
        elif msg.isBroadcast:
            if self.broadcastCallback:
                logger.info (f"Got broadcast with messagePath {msg.oscpath} and listOfArguments {msg.params}")
                self.broadcastCallback(self.ourPath, self.sessionName, self.ourClientNameUnderNSM, msg.oscpath, msg.params)
            else:
                logger.info (f"No callback for broadcast! Got messagePath {msg.oscpath} and listOfArguments {msg.params}")
        elif msg.oscpath == "/error":
            logger.warning("Got /error from NSM Server. Path: {} , Parameter: {}".format(msg.oscpath, msg.params))
        else:
            logger.warning("Reaction not implemented:. Path: {} , Parameter: {}".format(msg.oscpath, msg.params))


    def send(self, path:str, listOfParameters:list, host=None, port=None):
        """Send any osc message. Defaults to nsmd URL.
        Will not wait for an answer but return None."""
        if host and port:
            url = (host, port)
        else:
            url = self.nsmOSCUrl
        msg = _OutgoingMessage(path)
        for arg in listOfParameters:
            msg.add_arg(arg) #type is auto-determined by outgoing message
        self.sock.sendto(msg.build(), url)

    def getNsmOSCUrl(self):
        """Return and save the nsm osc url or raise an error"""
        nsmOSCUrl = getenv("NSM_URL")
        if not nsmOSCUrl:
            raise NSMNotRunningError("New-Session-Manager environment variable $NSM_URL not found.")
        else:
            #osc.udp://hostname:portnumber/
            o = urlparse(nsmOSCUrl)
            #return o.hostname, o.port #this always make the hostname lowercase. usually it does not matter, but we got crash reports. Alternative:
            return o.netloc.split(":")[0], o.port

    def getExecutableName(self):
        """Finding the actual executable name can be a bit hard
        in Python. NSM wants the real starting point, even if
        it was a bash script.
        """
        #TODO: I really don't know how to find out the name of the bash script
        fullPath = argv[0]
        assert os.path.dirname(fullPath) in os.environ["PATH"], (fullPath, os.path.dirname(fullPath), os.environ["PATH"]) #NSM requires the executable to be in the path. No excuses. This will never happen since the reference NSM server-GUI already checks for this.

        executableName = os.path.basename(fullPath)
        assert not "/" in executableName, executableName #see above.
        return executableName

    def announceOurselves(self):
        """Say hello to NSM and tell it we are ready to receive
        instructions

        /nsm/server/announce s:application_name s:capabilities s:executable_name i:api_version_major i:api_version_minor i:pid"""

        def buildClientFeaturesString():
            #:dirty:switch:progress:
            result = []
            if self.supportsSaveStatus:
                result.append("dirty")
            if self.hideGUICallback and self.showGUICallback:
                result.append("optional-gui")
            if result:
                return ":".join([""] + result + [""])
            else:
                return ""

        logger.info("Sending our NSM-announce message")

        announce = _OutgoingMessage("/nsm/server/announce")
        announce.add_arg(self.prettyName)  #s:application_name
        announce.add_arg(buildClientFeaturesString()) #s:capabilities
        announce.add_arg(self.executableName)  #s:executable_name
        announce.add_arg(1)  #i:api_version_major
        announce.add_arg(2)  #i:api_version_minor
        announce.add_arg(int(getpid())) #i:pid
        hostname, port = self.nsmOSCUrl
        assert hostname, self.nsmOSCUrl
        assert port, self.nsmOSCUrl
        self.sock.sendto(announce.build(), self.nsmOSCUrl)

        #Wait for /reply (aka 'Howdy, what took you so long?)
        data, addr = self.sock.recvfrom(1024)
        msg = _IncomingMessage(data)

        if msg.oscpath == "/error":
            originalMessage, errorCode, reason = msg.params
            logger.error("Code {}: {}".format(errorCode, reason))
            quit()

        elif msg.oscpath == "/reply":
            nsmAnnouncePath, welcomeMessage, managerName, self.serverFeatures = msg.params
            assert nsmAnnouncePath == "/nsm/server/announce", nsmAnnouncePath
            logger.info("Got /reply " + welcomeMessage)

            #Wait for /nsm/client/open
            data, addr = self.sock.recvfrom(1024)
            msg = _IncomingMessage(data)
            assert msg.oscpath == "/nsm/client/open", msg.oscpath
            self.ourPath, self.sessionName, self.ourClientNameUnderNSM = msg.params
            self.ourClientId = os.path.splitext(self.ourClientNameUnderNSM)[1][1:]
            logger.info("Got '/nsm/client/open' from NSM. Telling our client to load or create a file with name {}".format(self.ourPath))
            self.openOrNewCallback(self.ourPath, self.sessionName, self.ourClientNameUnderNSM) #Host function to either load an existing session or create a new one.
            logger.info("Our client should be done loading or creating the file {}".format(self.ourPath))
            replyToOpen = _OutgoingMessage("/reply")
            replyToOpen.add_arg("/nsm/client/open")
            replyToOpen.add_arg("{} is opened or created".format(self.prettyName))
            self.sock.sendto(replyToOpen.build(), self.nsmOSCUrl)
        else:
            raise ValueError("Unexpected message path after announce: {}".format((msg.oscpath, msg.params)))

    def announceGuiVisibility(self, isVisible):
        message = "/nsm/client/gui_is_shown" if isVisible else "/nsm/client/gui_is_hidden"
        self.isVisible = isVisible
        guiVisibility = _OutgoingMessage(message)
        logger.info("Telling NSM that our clients switched GUI visibility to: {}".format(message))
        self.sock.sendto(guiVisibility.build(), self.nsmOSCUrl)

    def announceSaveStatus(self, isClean):
        """Only send to the NSM Server if there was really a change"""
        if not self.supportsSaveStatus:
            return

        if not isClean == self.cachedSaveStatus:
            message = "/nsm/client/is_clean" if isClean else "/nsm/client/is_dirty"
            self.cachedSaveStatus = isClean
            saveStatus = _OutgoingMessage(message)
            logger.info("Telling NSM that our clients save state is now: {}".format(message))
            self.sock.sendto(saveStatus.build(), self.nsmOSCUrl)

    def _saveCallback(self, msg):
        logger.info("Telling our client to save as {}".format(self.ourPath))
        self.saveCallback(self.ourPath, self.sessionName, self.ourClientNameUnderNSM)
        replyToSave = _OutgoingMessage("/reply")
        replyToSave.add_arg("/nsm/client/save")
        replyToSave.add_arg("{} saved".format(self.prettyName))
        self.sock.sendto(replyToSave.build(), self.nsmOSCUrl)
        #it is assumed that after saving the state is clear
        self.announceSaveStatus(isClean = True)


    def _sessionIsLoadedCallback(self, msg):
        if self.sessionIsLoadedCallback:
            logger.info("Received 'Session is Loaded'. Our client supports it. Forwarding message...")
            self.sessionIsLoadedCallback()
        else:
            logger.info("Received 'Session is Loaded'. Our client does not support it, which is the default. Discarding message...")

    def sigtermHandler(self, signal, frame):
        """Wait for the user to quit the program

        The user function does not need to exit itself.
        Just shutdown audio engines etc.

        It is possible, that the client does not implement quit
        properly. In that case NSM protocol demands that we quit anyway.
        No excuses.

        Achtung GDB! If you run your program with
            gdb --args python foo.py
        the Python signal handler will not work. This has nothing to do with this library.
        """
        logger.info("Telling our client to quit.")
        self.exitProgramCallback(self.ourPath, self.sessionName, self.ourClientNameUnderNSM)
        #There is a chance that exitProgramCallback will hang and the program won't quit. However, this is broken design and bad programming. We COULD place a timeout here and just kill after 10s or so, but that would make quitting our responsibility and fixing a broken thing.
        #If we reach this point we have reached the point of no return. Say goodbye.
        logger.warning("Client did not quit on its own. Sending SIGKILL.")
        kill(getpid(), SIGKILL)
        logger.error("SIGKILL did nothing. Do it manually.")

    def debugResetDataAndExit(self):
        """This is solely meant for debugging and testing. The user way of action should be to
        remove the client from the session and add a new instance, which will get a different
        NSM-ID.
        Afterwards we perform a clean exit."""
        logger.warning("debugResetDataAndExit will now delete {} and then request an exit.".format(self.ourPath))
        if os.path.exists(self.ourPath):
            if os.path.isfile(self.ourPath):
                try:
                    os.remove(self.ourPath)
                except Exception as e:
                    logger.info(e)
            elif os.path.isdir(self.ourPath):
                try:
                    shutil.rmtree(self.ourPath)
                except Exception as e:
                    logger.info(e)
        else:
            logger.info("{} does not exist.".format(self.ourPath))
        self.serverSendExitToSelf()

    def serverSendExitToSelf(self):
        """If you want a very strict client you can block any non-NSM quit-attempts, like ignoring a
        qt closeEvent, and instead send the NSM Server a request to close this client.
        This method is a shortcut to do just that.
        """
        logger.info("Sending SIGTERM to ourselves to trigger the exit callback.")
        #if "server-control" in self.serverFeatures:
        #    message = _OutgoingMessage("/nsm/server/stop")
        #    message.add_arg("{}".format(self.ourClientId))
        #    self.sock.sendto(message.build(), self.nsmOSCUrl)
        #else:
        kill(getpid(), SIGTERM) #this calls the exit callback

    def serverSendSaveToSelf(self):
        """Some clients want to offer a manual Save function, mostly for psychological reasons.
        We offer a clean solution in calling this function which will trigger a round trip over the
        NSM server so our client thinks it received a Save instruction. This leads to a clean
        state with a good saveStatus and no required extra functionality in the client."""

        logger.info("instructing the NSM-Server to send Save to ourselves.")
        if "server-control" in self.serverFeatures:
            #message = _OutgoingMessage("/nsm/server/save") # "Save All" Command.
            message = _OutgoingMessage("/nsm/gui/client/save")
            message.add_arg("{}".format(self.ourClientId))
            self.sock.sendto(message.build(), self.nsmOSCUrl)
        else:
            logger.warning("...but the NSM-Server does not support server control. Server only supports: {}".format(self.serverFeatures))

    def changeLabel(self, label:str):
        """This function is implemented because it is provided by NSM. However, it does not much.
        The message gets received but is not saved.
        The official NSM GUI uses it but then does not save it.
        We would have to send it every startup ourselves.

        This is fine for us as clients, but you need to provide a GUI field to enter that label."""
        logger.info("Telling the NSM-Server that our label is now " + label)
        message = _OutgoingMessage("/nsm/client/label")
        message.add_arg(label)  #s:label
        self.sock.sendto(message.build(), self.nsmOSCUrl)

    def broadcast(self, path:str, arguments:list):
        """/nsm/server/broadcast s:path [arguments...]
        We, as sender, will not receive the broadcast back.

        Broadcasts starting with /nsm are not allowed and will get discarded by the server
        """
        if path.startswith("/nsm"):
            logger.warning("Attempted broadbast starting with /nsm. Not allwoed")
        else:
            logger.info("Sending broadcast " + path + repr(arguments))
            message = _OutgoingMessage("/nsm/server/broadcast")
            message.add_arg(path)
            for arg in arguments:
                message.add_arg(arg)  #type autodetect
            self.sock.sendto(message.build(), self.nsmOSCUrl)

    def importResource(self, filePath):
        """aka. import into session

        ATTENTION! You will still receive an absolute path from this function. You need to make
        sure yourself that this path will not be saved in your save file, but rather use a place-
        holder that gets replaced by the actual session path each time. A good point is after
        serialisation. search&replace for the session prefix ("ourPath") and replace it with a tag
        e.g. <sessionDirectory>. The opposite during load.
        Only such a behaviour will make your session portable.

        Do not use the following pattern: An alternative that comes to mind is to only work with
        relative paths and force your programs workdir to the session directory. Better work with
        absolute paths internally .

        Symlinks given path into session dir and returns the linked path relative to the ourPath.
        It can handles single files as well as whole directories.

        if filePath is already a symlink we do not follow it. os.path.realpath or os.readlink will
        not be used.

        Multilayer links may indicate a users ordering system that depends on
        abstractions. e.g. with mounted drives under different names which get symlinked to a
        reliable path.

        Basically do not question the type of our input filePath.

        tar with the follow symlink option has os.path.realpath behaviour and therefore is able
        to follow multiple levels of links anyway.

        A hardlink does not count as a link and will be detected and treated as real file.

        Cleaning up a session directory is either responsibility of the user
        or of our client program. We do not provide any means to unlink or delete files from the
        session directory.
        """

        #Even if the project was not saved yet now it is time to make our directory in the NSM dir.
        if not os.path.exists(self.ourPath):
            os.makedirs(self.ourPath)

        filePath = os.path.abspath(filePath) #includes normalisation
        if not os.path.exists(self.ourPath):raise FileNotFoundError(self.ourPath)
        if not os.path.isdir(self.ourPath): raise NotADirectoryError(self.ourPath)
        if not os.access(self.ourPath, os.W_OK): raise PermissionError("not writable", self.ourPath)

        if not os.path.exists(filePath):raise FileNotFoundError(filePath)
        if os.path.isdir(filePath): raise IsADirectoryError(filePath)
        if not os.access(filePath, os.R_OK): raise PermissionError("not readable", filePath)

        filePathInOurSession = os.path.commonprefix([filePath, self.ourPath]) == self.ourPath
        linkedPath = os.path.join(self.ourPath, os.path.basename(filePath))
        linkedPathAlreadyExists = os.path.exists(linkedPath)

        if not os.access(os.path.dirname(linkedPath), os.W_OK): raise PermissionError("not writable", os.path.dirname(linkedPath))


        if filePathInOurSession:
            #loadResource from our session dir. Portable session, manually copied beforehand or just loading a link again.
            linkedPath = filePath #we could return here, but we continue to get the tests below.
            logger.info(f"tried to import external resource {filePath} but this is already in our session directory. We use this file directly instead. ")

        elif linkedPathAlreadyExists and os.readlink(linkedPath) == filePath:
            #the imported file already exists as link in our session dir. We do not link it again but simply report the existing link.
            #We only check for the first target of the existing link and do not follow it through to a real file.
            #This way all user abstractions and file structures will be honored.
            linkedPath = linkedPath
            logger.info(f"tried to import external resource {filePath} but this was already linked to our session directory before. We use the old link: {linkedPath} ")

        elif linkedPathAlreadyExists:
            #A new file shall be imported but it would create a linked name which already exists in our session dir.
            #Because we already checked for a new link to the same file above this means actually linking a different file so we need to differentiate with a unique name
            firstpart, extension = os.path.splitext(linkedPath)
            uniqueLinkedPath = firstpart + "." + uuid4().hex + extension
            assert not os.path.exists(uniqueLinkedPath)
            os.symlink(filePath, uniqueLinkedPath)
            logger.info(self.ourClientNameUnderNSM + f":pysm2: tried to import external resource {filePath} but potential target link {linkedPath} already exists. Linked to {uniqueLinkedPath} instead.")
            linkedPath = uniqueLinkedPath

        else: #this is the "normal" case. External resources will be linked.
            assert not os.path.exists(linkedPath)
            os.symlink(filePath, linkedPath)
            logger.info(f"imported external resource {filePath} as link {linkedPath}")

        assert os.path.exists(linkedPath), linkedPath
        return linkedPath

class NullClient(object):
    """Use this as a drop-in replacement if your program has a mode without NSM but you don't want
    to change the code itself.
    This was originally written for programs that have a core-engine and normal mode of operations
    is a GUI with NSM but they also support commandline-scripts and batch processing.
    For these you don't want NSM."""

    def __init__(self, *args, **kwargs):
        self.realClient = False
        self.ourClientNameUnderNSM = "NSM Null Client"

    def announceSaveStatus(self, *args):
        pass

    def announceGuiVisibility(self, *args):
        pass

    def reactToMessage(self):
        pass

    def importResource(self):
        return ""

    def serverSendExitToSelf(self):
        quit()



# This file is part of jack_mixer
#
# Copyright (C) 2006-2009 Nedko Arnaudov <nedko@arnaudov.name>
# Copyright (C) 2009-2020 Frederic Peters <fpeters@0d.be> et al.
# Copyright (C) 2020-2021 Christopher Arndt <info@chrisarndt>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from random import random

import gi  # noqa: F401
from gi.repository import Gtk
from gi.repository import Gdk


# CSS widget styling
DEFAULT_CSS = """\
/* Global color definitions */

@define-color monitor_bgcolor_hover #C0BFBC;
@define-color monitor_bgcolor_checked #9A9996;
@define-color mute_bgcolor_hover #FFBC80;
@define-color mute_bgcolor_checked #FF7800;
@define-color solo_bgcolor_hover #76A28E;
@define-color solo_bgcolor_checked #26A269;
@define-color prefader_bgcolor_hover #A6C2E4;
@define-color prefader_bgcolor_checked #3584E4;


/* Channel strips */

.top_label {
    padding: 0px .1em;
    min-height: 1.5rem;
}

.wide {
    font-size: medium;
}

.narrow {
    font-size: smaller;
}

.vbox_fader {
    border: 1px inset #111;
}

.readout {
    font-size: 80%;
    margin: .1em;
    padding: 0;
    border: 1px inset #111;
    color: white;
    background-color: #333;
    background-image: none;
}


/* Channel buttons */

button {
    padding: 0px .2em;
}
button.prefader_meter {
    font-size: smaller;
}
button.monitor:hover,
button.mute:hover,
button.solo:hover,
button.prefader:hover,
button.prefader_meter:hover,
button.monitor:checked,
button.mute:checked,
button.solo:checked,
button.prefader:checked,
button.prefader_meter:checked {
    color: white;
    text-shadow: unset;
    background-image: none;
}
button.monitor:hover {
    background-color: @monitor_bgcolor_hover;
}
button.monitor:checked {
    background-color: @monitor_bgcolor_checked;
}
button.mute:hover {
    background-color: @mute_bgcolor_hover;
}
button.mute:checked {
    background-color: @mute_bgcolor_checked;
}
button.solo:hover {
    background-color: @solo_bgcolor_hover;
}
button.solo:checked {
    background-color: @solo_bgcolor_checked;
}
button.prefader:hover {
    background-color: @prefader_bgcolor_hover;
}
button.prefader:checked {
    background-color: @prefader_bgcolor_checked;
}
button.prefader_meter:hover {
    background-color: @prefader_bgcolor_hover;
}
button.prefader_meter:checked {
    background-color: @prefader_bgcolor_checked;
}


/* Control groups */

.control_group {
    min-width: 0px;
    padding: 0px;
}

.control_group .label,
.control_group .mute,
.control_group .prefader,
.control_group .solo {
    font-size: smaller;
    padding: 0px .1em;
}

.control_group .mute:hover,
.control_group .solo:hover,
.control_group .prefader:hover,
.control_group .mute:checked,
.control_group .solo:checked,
.control_group .prefader:checked {
    color: white;
    text-shadow: unset;
    background-image: none;
}
.control_group .mute:hover {
    background-color: @mute_bgcolor_hover;
}
.control_group .mute:checked {
    background-color:@mute_bgcolor_checked;
}
.control_group .solo:hover {
    background-color: @solo_bgcolor_hover;
}
.control_group .solo:checked {
    background-color: @solo_bgcolor_checked;
}
.control_group .prefader:hover {
    background-color: @prefader_bgcolor_hover;
}
.control_group .prefader:checked {
    background-color: @prefader_bgcolor_checked;
}


/* Peak meters */

.over_zero {
    background-color: #cc4c00;
}

.is_nan {
    background-color: #b20000;
}
"""

COLOR_TMPL_CSS = """\
.{} {{
    background-color: {};
    color: {};
}}
"""


def add_css_provider(css, priority=Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION):
    css_provider = Gtk.CssProvider()
    css_provider.load_from_data(css.encode("utf-8"))
    context = Gtk.StyleContext()
    screen = Gdk.Screen.get_default()
    context.add_provider_for_screen(screen, css_provider, priority)


def get_text_color(background_color):
    """Calculate the luminance of the given color (GdkRGBA)
    and return an appropriate text color."""
    # luminance coefficients taken from section C-9 from
    # http://www.faqs.org/faqs/graphics/colorspace-faq/
    brightess = (
        background_color.red * 0.212671
        + background_color.green * 0.715160
        + background_color.blue * 0.072169
    )

    if brightess > 0.5:
        return "black"
    else:
        return "white"


def load_css_styles():
    add_css_provider(DEFAULT_CSS)


def set_background_color(widget, name, color):
    color_string = color.to_string()
    add_css_provider(COLOR_TMPL_CSS.format(name, color_string, get_text_color(color)))
    widget_context = widget.get_style_context()
    widget_context.add_class(name)


def random_color():
    return Gdk.RGBA(random(), random(), random(), 1)
